﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using Luticate2.Auth.Utils.Business.Converters;
using Luticate2.Auth.Utils.Business.Converters.ExpressionConverter;
using Luticate2.Auth.Utils.Dbo;
using Luticate2.Auth.Utils.Dbo.Fields;
using Luticate2.Auth.Utils.Dbo.Pagination;
using Luticate2.Auth.Utils.Dbo.Result;
using Luticate2.Auth.Utils.Interfaces;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.DependencyInjection;

namespace Luticate2.Auth.Utils.DataAccess.Crud
{
    public class LuEfCrudDataAccess<TDbo, TModel, TDbContext> : ILuCrud<TDbo>
        where TModel : class
        where TDbContext : DbContext
        where TDbo : class
    {
        protected IServiceProvider ServiceProvider { get; }
        protected ILuConvertersAllocator ConvertersAllocator { get; }
        protected ILuConvertersTypeConverter ConvertersTypeConverter { get; }

        public LuEfCrudDataAccess(IServiceProvider serviceProvider)
        {
            ServiceProvider = serviceProvider;
            ConvertersAllocator = ServiceProvider.GetRequiredService<ILuConvertersAllocator>(); // TODO Make a DbContext or DataAccess specific ILuConvertersAllocator (in case of multiple DbContext)
            ConvertersTypeConverter = ServiceProvider.GetRequiredService<ILuConvertersTypeConverter>(); // TODO same as above
        }

        protected virtual LuConvertersOptions GetOptions()
        {
            return new LuConvertersOptions
            {
                Allocator = ConvertersAllocator,
                Parameters = new Dictionary<ParameterExpression, Expression>(),
                TypeConverter = ConvertersTypeConverter
            };
        }

        protected virtual LuResult<TModel> ConvertDboToModel(LuPartialFieldsDbo partialInput, TDbo dbo)
        {
            var copier = ServiceProvider.GetService<ILuObjectConverter<TDbo, TModel>>();
            if (copier == null)
            {
                return LuResult<TModel>.Error(LuStatus.InternalError.ToInt(),
                    $"Could not get service: {nameof(ILuObjectConverter)}<{typeof(TDbo).Name}, {typeof(TModel).Name}>");
            }

            var options = GetOptions();
            var dstObjResult = options.Allocator.GetInstance(typeof(TModel));
            if (!dstObjResult)
            {
                return dstObjResult.To<TModel>();
            }
            var dstObj = dstObjResult.Data;
            var result = copier.Convert(dbo, dstObj, new LuFieldDbo(), partialInput, GetOptions());
            return result.Select(o => o as TModel);
        }

        protected virtual LuResult<TDbo> ConvertModelToDbo(LuPartialFieldsDbo partialResponse, TModel model)
        {
            var copier = ServiceProvider.GetService<ILuObjectConverter<TModel, TDbo>>();
            if (copier == null)
            {
                return LuResult<TDbo>.Error(LuStatus.InternalError.ToInt(),
                    $"Could not get service: {nameof(ILuObjectConverter)}<{typeof(TModel).Name}, {typeof(TDbo).Name}>");
            }

            var options = GetOptions();
            var dstObjResult = options.Allocator.GetInstance(typeof(TDbo));
            if (!dstObjResult)
            {
                return dstObjResult.To<TDbo>();
            }
            var dstObj = dstObjResult.Data;

            var result = copier.Convert(model, dstObj, new LuFieldDbo(), partialResponse, GetOptions());
            return result.Select(o => o as TDbo);
        }

        protected virtual LuResult<T> HandleError<T>(Exception e)
        {
            return LuResult<T>.Error(LuStatus.DbError.ToInt(), e);
        }

        protected LuResult<T> Execute<T>(Func<TDbContext, DbSet<TModel>, LuResult<T>> action)
        {
            try
            {
                using (var db = ServiceProvider.GetService<TDbContext>())
                {
                    return action(db, db.Set<TModel>());
                }
            }
            catch (Exception e)
            {
                var result = HandleError<T>(e);
                return result;
            }
        }

        protected virtual LuResult<IQueryable<TModel>> Include(LuPartialFieldsDbo partialResponse, IQueryable<TModel> queryable)
        {
            return LuResult<IQueryable<TModel>>.Ok(queryable);
        }

        protected virtual LuResult<IQueryable<TModel>> Filter(LuFilterDbo filter, IQueryable<TModel> queryable)
        {
            if (filter.Expression == null)
            {
                return LuResult<IQueryable<TModel>>.Ok(queryable);
            }

            var lambdaDbo = (Expression<Func<TDbo, bool>>) filter.Expression;

            var options = GetOptions();
            var converter = new LuExpressionConverterVisitor(options, ServiceProvider);
            var lambdaModel = converter.Visit(lambdaDbo) as Expression<Func<TModel, bool>>;// TODO Handle errors
            return LuResult<IQueryable<TModel>>.Ok(queryable.Where(lambdaModel));
        }

        protected virtual LuResult<IQueryable<TModel>> OrderByField(LuOrderByFieldDbo orderByField, IQueryable<TModel> queryable)
        {
            var lambdaDbo = (Expression<Func<TDbo, object>>) orderByField.Expression;

            var options = GetOptions();
            var converter = new LuExpressionConverterVisitor(options, ServiceProvider);
            var lambdaModelNoConvert = converter.Visit(lambdaDbo) as LambdaExpression;// TODO Handle errors
            var lambdaModel = Expression.Lambda<Func<TModel, object>>(lambdaModelNoConvert.Body, lambdaModelNoConvert.Parameters);

            IOrderedQueryable<TModel> ordered;

            if (queryable is IOrderedQueryable<TModel> orderedQueryable)
            {
                if (orderByField.Asc)
                {
                    ordered = orderedQueryable.ThenBy(lambdaModel);
                }
                else
                {
                    ordered = orderedQueryable.ThenByDescending(lambdaModel);
                }
            }
            else
            {
                if (orderByField.Asc)
                {
                    ordered = queryable.OrderBy(lambdaModel);
                }
                else
                {
                    ordered = queryable.OrderByDescending(lambdaModel);
                }
            }

            return LuResult<IQueryable<TModel>>.Ok(ordered);
        }

        protected virtual LuResult<IQueryable<TModel>> OrderBy(LuOrderByDbo orderBy, IQueryable<TModel> queryable)
        {
            var ordered = queryable;
            foreach (var field in orderBy.OrderByFields)
            {
                var orderedResult = OrderByField(field, ordered);
                if (!orderedResult)
                {
                    return orderedResult.To<IQueryable<TModel>>();
                }

                ordered = orderedResult.Data;
            }

            return LuResult<IQueryable<TModel>>.Ok(ordered ?? queryable);
        }

        protected virtual LuResult<IQueryable<TModel>> Paginate(int page, int perPage, IQueryable<TModel> queryable)
        {
            var paginated = queryable.Skip(page * perPage).Take(perPage);
            return LuResult<IQueryable<TModel>>.Ok(paginated);
        }

        public virtual LuResult<IList<TDbo>> Create(LuPartialFieldsDbo partialResponse, LuPartialFieldsDbo partialInput, IEnumerable<TDbo> dbos)
        {
            throw new NotImplementedException();
        }

        public virtual LuResult<LuPaginatedDbo<TDbo>> Read(LuPartialFieldsDbo partialResponse, LuPaginatedParamsDbo paginationParams)
        {
            var result = Execute((context, set) =>
            {
                var includeResult = Include(partialResponse, set);
                if (!includeResult)
                {
                    return includeResult.To<LuPaginatedDbo<TDbo>>();
                }
                var included = includeResult.Data;

                var orderByResult = OrderBy(paginationParams.OrderBy, included);
                if (!orderByResult)
                {
                    return orderByResult.To<LuPaginatedDbo<TDbo>>();
                }
                var ordered = orderByResult.Data;
                
                var filteredResult = Filter(paginationParams.Filter, ordered);
                if (!filteredResult)
                {
                    return filteredResult.To<LuPaginatedDbo<TDbo>>();
                }
                var filtered = filteredResult.Data;
                var count = filtered.Count();

                var paginatedResult = Paginate(paginationParams.Page, paginationParams.PerPage, filtered);
                if (!paginatedResult)
                {
                    return paginatedResult.To<LuPaginatedDbo<TDbo>>();
                }
                var paginated = paginatedResult.Data;

                var dboList = new List<TDbo>();
                foreach (var model in paginated)
                {
                    var dboResult = ConvertModelToDbo(partialResponse, model);
                    if (!dboResult)
                    {
                        return dboResult.To<LuPaginatedDbo<TDbo>>();
                    }
                    var dbo = dboResult.Data;

                    dboList.Add(dbo);
                }

                return LuResult<LuPaginatedDbo<TDbo>>.Ok(new LuPaginatedDbo<TDbo>
                {
                    Count = count,
                    Data = dboList
                });
            });
            return result;
        }

        public virtual LuResult<IList<TDbo>> Update(LuPartialFieldsDbo partialResponse, LuPartialFieldsDbo partialInput, IEnumerable<TDbo> dbos)
        {
            throw new NotImplementedException();
        }

        public virtual LuResult<IList<TDbo>> Delete(LuPartialFieldsDbo partialResponse, LuPartialFieldsDbo partialInput, IEnumerable<TDbo> dbos)
        {
            throw new NotImplementedException();
        }
    }
}